/*
 * File : scheduler.C
 *
 * Description : This file contains method definitions for the
 *	scheduler class.
 */

#include <stdio.h>
#include <libc.h>
#include <sys/types.h>
#include <sys/time.h>
#include "scheduler.h"
#include "eventHandler.h"
#include "trace.h"

DEFINE_TRACE_MODULE(schedulerModule,"schedulerModule");

// Init static data
scheduler * scheduler::_scheduler = 0;

/*
 * Function : scheduler()
 *
 * Description : This is the constructor. Initializes data.
 */

scheduler::scheduler()
{
	_eventHandlerList = new eventHandlerList;
	_scheduler = this;
}

/*
 * Function : ~scheduler()
 *
 * Description : This is the destructor. Frees memory.
 */


scheduler::~scheduler()
{
	if ( _eventHandlerList )
		delete _eventHandlerList;
}

/*
 * Function : Run
 *
 * Description : This is the main event loop. On every event
 *	handler being scheduled, it first calls SetEvent. This
 *	causes the event handler to either set the file 
 *	descriptor sets for the select call, or specify a timeout
 *	value. The lowest timeout value among all the event
 *	handlers is chosen for the select system call.
 *	Once the select call returns, it calls CheckEvent() on
 *	each event handler. The event handlers will check for
 *	input on files, or timeouts and take appropriate action.
 */

void 		
scheduler::Run()
{
	TRACE(schedulerModule,1,("Entering Run\n"));

	while ( 1 ) { // Infinite loop....

		fd_set rset, wset, eset;

		// initialize file descriptor sets.
		FD_ZERO(&rset); 
		FD_ZERO(&wset);
		FD_ZERO(&eset);

		timeval timeout;
		timeout.tv_sec = timeout.tv_usec = -1;

		eventHandler *es = _eventHandlerList->Head();
		timeval temp;
		temp.tv_sec = -1;

		/*
		 * For every event handler, call SetEvent
		 * Chose lowest timeout value.
		 */

		while ( es != 0 ) {
			es->SetEvent(&rset,&wset,&eset,&temp);
			if ( temp.tv_sec >= 0 && 
				( timeout.tv_sec == -1 || temp.tv_sec < timeout.tv_sec )) {
					timeout.tv_sec = temp.tv_sec;
					timeout.tv_usec = 0;
				}

			es = es->Next();
		}

		TRACE(schedulerModule,1,("Timeout : %d\n",timeout.tv_sec));

		// If timeout.tv_sec is still -1, then it means there is
		// no timeout needed. Wait infinitely for a read/write/exception

		if ( timeout.tv_sec == -1 ) {
			if ( select (10, &rset, &wset, &eset, 0) == -1 ) {
				perror("Error calling select");
			}
		}
		else {
			if ( select (10, &rset, &wset, &eset, &timeout) == -1 ) {
				perror("Error calling select");
			}
		}

		/*
		 * Now, run through list of event handlers again, and
		 * call CheckEvent on each.
		 */

		es = _eventHandlerList->Head();
		while ( es != 0 ) {

			struct timezone t;
			timeval curtime;

			(void)gettimeofday(&curtime,&t);
			if ( curtime.tv_usec >= 500000 )
				curtime.tv_sec++;
			es->CheckEvent(&rset,&wset,&eset,&curtime);

			es = es->Next();
		}
		
		/*
		 * Some event handlers might have marked themselves up
		 * for deletion - call CleanUpMarked on list to delete
		 * them.
		 */

		_eventHandlerList->CleanUpMarked();
	}
}
/*
 * Function : AddEventHandler
 *
 * Description : Adds event handler to be scheduled.
 */

void
scheduler::AddEventHandler(eventHandler *hdl)
{
	_eventHandlerList->Append(hdl);
}
/*
 * Function : DeleteEventHandler()
 *
 * Description : Deletes event handler from list of handlers
 *	to be scheduled.
 */

void 		
scheduler::DeleteEventHandler(eventHandler *hdl)
{
	_eventHandlerList->Remove(hdl);
}
/*
 * Function : NumOfEventHandlers()
 *
 * Description : Returns number of event handlers being scheduled.
 */

int			
scheduler::NumOfEventHandlers()
{
	return _eventHandlerList->Length();
}

