/*
 * File : window.C
 *
 * Description : This file contains the method descriptions for
 * the window class.
 */

#include "types.h"
#include "trace.h"
#include "window.h"


DEFINE_TRACE_MODULE(windowModule,"windowModule");

/* 
 * Function : window()
 *
 * Description : This is the constructor. It initializes the
 *	variables.
 */

window::window(int modulus) :
_modulus(modulus),
_vs(0), _vr(0),
_ackPending(0)
{
}

/* 
 * Function : ~window()
 *
 * Description : Destructor. Empties the queue.
 */

window::~window()
{
	// Remove all frames from the queue.
	UnqueueFrames(_xmitList.Length());
}

/* 
 * Function : GetNextToXmit()
 *
 * Description : This function formats an I-frame and returns
 *	a pointer to the dataPacket portion of the frame object.
 * This is the next in queue I-frame.
 * If it returns 0, it means there are no more I-frames
 * that can be transmitted at this time. Either there are
 * none, or the window is full, or acks are outstanding
 * on all of them.
 */


dataPacket *
window::GetNextToXmit()
{
	TRACE(windowModule,1,("Entering GetNextToXmit\n"));

	if ( _ackPending == K ) {
		// Window is FULL. No more I-frames can be transmitted.
		TRACE(windowModule,2,("Window FULL\n"));
		return 0; // Maximum outstanding acks
	}

	if ( _ackPending == _xmitList.Length() ) {
		// All have been transmitted, but not acknowledged.
		TRACE(windowModule,2,("All transmitted!\n"));
		return 0; // All outstanding acks
	}

	// Get next in list.
	frame *f = _xmitList.GetNthMember(_ackPending);
	TRACE(windowModule,1,("Extracted %dth frame : %x\n",_ackPending,f));
	int pf;

	// Check if we need to set the P bit to 1 or 0.

	if ( _ackPending == (K-1)) 
		// Its the last one before the window gets full.
		pf = 1;

	else if ( (_xmitList.Length() - _ackPending ) == 1) 
		// Its the last one in the list.
		pf = 1;

	else
		pf = 0;

	// Form the I-frame.
	f->SetParams(IFRAME,pf,_vr,_vs);
	f->FormFrame(TRUE);
	TRACE(windowModule,1,("Xmitting Ns : %d Nr : %d\n",_vs,_vr));
		

	// Increment V(s) state variable.
	_vs = (_vs + 1) % _modulus; // Modulo increment

	// Increment unacked count.
	_ackPending++;

	// Return the dataPacket pointer.
	return (f->GetDataPacket());
}

/* 
 * Function : ProcessAck()
 *
 * Description : This function processes acknowledgements in the
 *	form of a received N(r). It figures out if the N(r) is
 *	valid. If it is valid, and it is actually acknowledging
 *	some I-frames, then those I-frames are removed from
 *	the queue.
 * Returns -1 if N(r) is invalid, else returns the number
 *	of I-frames acked.
 */

int 
window::ProcessAck(int nr)
{
	TRACE(windowModule,1,("In ProcessAck, _ackPending : %d _vs = %d nr = %d\n",
			_ackPending,_vs,nr)); 

	if ( nr == _vs )  { 
		// All Acked, remove all Iframes from queue.

		UnqueueFrames(_ackPending);
		int temp = _ackPending;
		_ackPending = 0;
		return temp;
	}


	// Check if N(r) is valid.
	int i = 1;
	int temp = 0;

	while ( i <= _ackPending ) {
		
		temp =  ( _vs >= i ) ? 
				(_vs - i)%_modulus : 
				(_modulus - i + _vs );

		TRACE(windowModule,2,("Checking for : %d %d\n",temp,nr)); 
		if ( nr == temp )
			break;
		else
			i++;
	} 
	if ( i > _ackPending ) // Wrong N(r)!
		return -1;

	// Valid NR
	int acked = _ackPending - i;
	_ackPending = i;

	TRACE(windowModule,2,("Acked: %d pending %d\n",acked,_ackPending));

	UnqueueFrames(acked); // remove acked Iframes from queue
	return acked; // Return number of I-frames acknowledged.
}

/* 
 * Function : AddFrame()
 *
 * Description : Adds an I-frame to the queue of Iframes
 * to be transmitted.
 */

void 
window::AddFrame(dataPacket *dp)
{
	// Create a frame.
	TRACE(windowModule,1,("Entering AddFrame Length : %d\n",
		_xmitList.Length()));

	frame *f = new frame(dp,TRUE);

	// And add to (FIFO)list.
	_xmitList.Prepend(f);

	TRACE(windowModule,1,("Added Frame. Len : %d Frame : %x\n",
		_xmitList.Length(),f));
}


/* 
 * Function : CheckSequence()
 *
 * Description : Checks if the received N(s) value
 *	of an I-frame is in sequence. Return TRUE
 * if it is, else FALSE. Increments state 
 * variable V(s) if valid N(s) was received.
 */


Boolean 
window::CheckSequence(int ns)
{
	if ( ns != _vr )
		return FALSE;

	_vr = (_vr+1)%_modulus;
	return TRUE;
}

/* 
 * Function : UnqueueFrames()
 *
 * Description : Removes the indicated number of 
 * Iframes from the _xmitList. 
 */

void
window::UnqueueFrames(int num)
{
	int i= 0;
	TRACE(windowModule,2,("Len of list: %d\n",_xmitList.Length()));
	if ( num > _xmitList.Length() ) {
		TRACE(windowModule,2,("NUM > LENGTH!!\n"));
		return;
	}
	TRACE(windowModule,2,("Removing IFRAMEs: "));

	while ( i < num ) {

		TRACE(windowModule,2,("%d ",i));
		frame *f = _xmitList.Top();
		if ( f != 0 )
			_xmitList.Remove(f);
		i++;
	}
	TRACE(windowModule,2,("\nLen of list: %d\n",_xmitList.Length()));
}

/* 
 * Function : Reset()
 *
 * Description : Resets the window by reinitializing
 * all state variables, and discarding any I-frames
 * in the queue.
 */

void 
window::Reset()
{

	TRACE(windowModule,2,("Resetting Window\n"));
	_vr = _vs = 0;	
	if ( _xmitList.Length() > 0 )
		UnqueueFrames(_xmitList.Length());
	_ackPending = 0;
}

/* 
 * Function : StartRetransmission()
 *
 * Description : Called when the data-link wants
 * to retransmit I-frames. The send state 
 * variable V(s) needs to be reset, as does
 * _ackPending.
 */

void
window::StartRetransmission()
{
	// Start of retransmission cycle
	if ( _ackPending == 0 )
		return;

	// Else we need to modify v(s), which will be the
	// sequence number for the next I-frame to be
	// transmitted.

	if ( _vs > _ackPending )
		_vs = _vs - _ackPending;
	else
		_vs = ( _modulus - _ackPending ) + _vs;

	_ackPending = 0;
}

